import argparse
import os
import sys

import src.config
from src.data_manipulation import logger
import src.execute_anonymization as execute_anonymization
import src.execute_pipeline_processing as execute_pipeline_processing
import src.pipelines as pipelines
import src.start as start
from src.statistics import execute_statistics, execute_original_vols_statistics
from src.structures import AnonymizationMethod, \
    ProcessingPipeline, AnonymizationSetting, ORIGINAL, ANONYMIZATION_METHODS


def validate_setting(a_method: AnonymizationMethod,
                     a_setting: AnonymizationSetting):
    """Checks if a_method supports chosen a_setting.

    Args:
        a_method (AnonymizationMethod): anonymization method
        a_setting (AnonymizationSetting): setting for anonymization method

    Raises:
        ValueError: If a_method doesn't support a_setting
    """
    if (a_setting == AnonymizationSetting.DEFAULT):
        return

    # check if method has setting in commands
    for _, options in a_method.settings.items():
        if a_setting in options:
            return

    logger.exception((
        f'Incorrect setting {str(a_setting)} for {str(a_method)}.'
        f' Valid settings include: {str(a_method.settings)}'))
    raise ValueError((
        f'Incorrect setting {str(a_setting)} for {str(a_method)}.'
        f' Valid settings include: {str(a_method.settings)}'))


def init_start(_: argparse.Namespace):
    ''' Calls function which initializes working directory '''
    logger.info('Start initialized')
    start.init_working_dir()


def print_subjects(_: argparse.Namespace) -> None:
    ''' Prints all loaded subjects '''
    print(f'SUBJECTS: {src.config.SUBJECTS}')


def get_setting(args: argparse.Namespace) -> AnonymizationSetting:
    if (args.cost):
        return args.cost
    elif (args.afni_mask):
        return AnonymizationSetting.BRAIN_MASK_AFNI
    elif (args.bias_correct):
        return AnonymizationSetting.BIAS_CORRECTED

    return AnonymizationSetting.DEFAULT


def anonymize(args):
    ''' Calls function which executes anonymization '''
    method = ANONYMIZATION_METHODS.get(args.method)
    a_setting = get_setting(args)

    validate_setting(method, a_setting)

    execute_anonymization.execute_anonymization(
        args.subjects,
        method,
        a_setting
    )


def execute_pipeline(args: argparse.Namespace, pipeline: ProcessingPipeline):
    """Calls function which executes SPM or fMRIPrep. Validates inputs.

    Args:
        args (argparse.Namespace): args of argparse

    Raises:
        ValueError: if anonymization setting is not supported for method
    """
    if args.method == 'original':
        if (args.cost or args.bias_correct or args.afni_mask):
            logger.exception(
                'Original image can\'t have anonymization setting')
            raise ValueError(
                'Original image can\'t have anonymization setting')
        method = ORIGINAL
        a_setting = None
    else:
        method = ANONYMIZATION_METHODS.get(args.method)
        a_setting = get_setting(args)
        validate_setting(method, a_setting)

    if (pipeline == ProcessingPipeline.FMRIPREP):
        execute_pipeline_processing.fmriprep_anat_pipeline(
                            args.subjects, method, a_setting)
    else:
        if (args.original_warp):
            execute_pipeline_processing.spm_normalization_from_original_warp(
                            args.subjects, method, a_setting)
        else:
            execute_pipeline_processing.spm_normalization(
                            args.subjects, method, a_setting)


def spm_normalization(args: argparse.Namespace):
    ''' Calls pipeline execution set to SPM '''
    execute_pipeline(args, ProcessingPipeline.SPM)


def fmriprep_anat_pipeline(args: argparse.Namespace):
    ''' Calls pipeline execution set to fMRIPrep '''
    execute_pipeline(args, ProcessingPipeline.FMRIPREP)


def statistics(args: argparse.Namespace):
    methods = [ANONYMIZATION_METHODS[key] for key in args.methods]
    a_setting = get_setting(args)
    for method in methods:
        validate_setting(method, a_setting)
    logger.info('Statistics initialized')
    execute_statistics(args.subjects, methods, a_setting, args.pipeline)


def statistics_pipeline(args: argparse.Namespace):
    methods = [ANONYMIZATION_METHODS[key] for key in args.methods]
    a_setting = get_setting(args)
    for method in methods:
        validate_setting(method, a_setting)
    logger.info('Statistics pipeline')
    pipelines.statistics_pipeline(args.subjects, methods,
                                  a_setting, args.pipeline, args.skip_original)


def cost_statistics_pipeline(args: argparse.Namespace):
    method = ANONYMIZATION_METHODS.get(args.method)
    logger.info('Cost statistics pipeline')
    pipelines.cost_statistics_pipeline(
                args.subjects, method, args.pipeline,
                args.skip_default_cost, args.skip_original)


def original_vols(args: argparse.Namespace):
    logger.info('Per voxel volumes of original images')
    execute_original_vols_statistics(args.subjects, args.pipeline)


def configure_parsers():
    ''' Initializes all command parsers for application '''

    # parser initialization
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    # subparsers initialization
    parser_start = subparsers.add_parser('start')
    parser_start.set_defaults(func=init_start)

    parser_subjects = subparsers.add_parser('subjects')
    parser_subjects.set_defaults(func=print_subjects)

    parser_original_vols = subparsers.add_parser('orig_vols')
    parser_original_vols.add_argument(
                        '-s', '--subjects',
                        type=str,
                        choices=src.config.SUBJECTS,
                        nargs='+',
                        default=src.config.SUBJECTS)
    parser_original_vols.add_argument(
                        '-p', '--pipeline',
                        required=True,
                        type=ProcessingPipeline.from_string,
                        choices=list(ProcessingPipeline))
    parser_original_vols.set_defaults(func=original_vols)

    parser_anonymize = subparsers.add_parser('anonymize')
    parser_anonymize.add_argument(
                        '-m', '--method',
                        required=True,
                        choices=ANONYMIZATION_METHODS.keys())
    parser_anonymize.add_argument(
                        '-s', '--subjects',
                        type=str,
                        choices=src.config.SUBJECTS,
                        nargs='+',
                        default=src.config.SUBJECTS)
    group_anonymize = parser_anonymize.add_mutually_exclusive_group()
    group_anonymize.add_argument(
                        '-c', '--cost',
                        type=AnonymizationSetting.from_string,
                        choices=(ANONYMIZATION_METHODS['pydeface']
                                 .settings['cost']
                                 .union(
                                   ANONYMIZATION_METHODS['afni_refacer_deface']
                                   .settings['cost'])))
    group_anonymize.add_argument('-B', '--bias_correct', action='store_true')
    group_anonymize.add_argument('-a', '--afni_mask', action='store_true')
    parser_anonymize.set_defaults(func=anonymize)

    parser_spm = subparsers.add_parser('spm')
    parser_spm.add_argument(
                        '-s', '--subjects',
                        type=str,
                        choices=src.config.SUBJECTS,
                        nargs='+',
                        default=src.config.SUBJECTS)
    parser_spm.add_argument(
                        '-m', '--method',
                        required=True,
                        type=str,
                        choices=list(
                            ANONYMIZATION_METHODS.keys()) + ['original'])
    group_spm = parser_spm.add_mutually_exclusive_group()
    group_spm.add_argument(
                        '-c', '--cost',
                        type=AnonymizationSetting.from_string,
                        choices=(ANONYMIZATION_METHODS['pydeface']
                                 .settings['cost']
                                 .union(
                                   ANONYMIZATION_METHODS['afni_refacer_deface']
                                   .settings['cost'])))
    group_spm.add_argument('-B', '--bias_correct', action='store_true')
    group_spm.add_argument('-a', '--afni_mask', action='store_true')
    parser_spm.add_argument('-orig', '--original_warp', action='store_true')
    parser_spm.set_defaults(func=spm_normalization)

    parser_fmriprep = subparsers.add_parser('fmriprep')
    parser_fmriprep.add_argument(
                            '-s', '--subjects',
                            type=str,
                            choices=src.config.SUBJECTS,
                            nargs='+',
                            default=src.config.SUBJECTS)
    parser_fmriprep.add_argument(
                            '-m', '--method',
                            required=True,
                            type=str,
                            choices=list(
                                ANONYMIZATION_METHODS.keys()) + ['original'])
    group_fmriprep = parser_fmriprep.add_mutually_exclusive_group()
    group_fmriprep.add_argument(
                        '-c', '--cost',
                        type=AnonymizationSetting.from_string,
                        choices=(ANONYMIZATION_METHODS['pydeface']
                                 .settings['cost']
                                 .union(
                                   ANONYMIZATION_METHODS['afni_refacer_deface']
                                   .settings['cost'])))
    group_fmriprep.add_argument('-B', '--bias_correct', action='store_true')
    group_fmriprep.add_argument('-a', '--afni_mask', action='store_true')
    parser_fmriprep.set_defaults(func=fmriprep_anat_pipeline)

    parser_statistics = subparsers.add_parser(
                            'statistics')
    parser_statistics.add_argument(
                            '-m', '--methods',
                            type=str,
                            choices=list(ANONYMIZATION_METHODS.keys()),
                            nargs='+',
                            default=list(ANONYMIZATION_METHODS.keys()))
    parser_statistics.add_argument(
                            '-s', '--subjects', type=str,
                            choices=src.config.SUBJECTS, nargs='+',
                            default=src.config.SUBJECTS)
    parser_statistics.add_argument(
                            '-p', '--pipeline',
                            required=True,
                            type=ProcessingPipeline.from_string,
                            choices=list(ProcessingPipeline))
    group_statistics = parser_statistics.add_mutually_exclusive_group()
    group_statistics.add_argument(
                        '-c', '--cost',
                        type=AnonymizationSetting.from_string,
                        choices=(ANONYMIZATION_METHODS['pydeface']
                                 .settings['cost']
                                 .union(
                                   ANONYMIZATION_METHODS['afni_refacer_deface']
                                   .settings['cost'])))
    group_statistics.add_argument('-B', '--bias_correct', action='store_true')
    group_statistics.add_argument('-a', '--afni_mask', action='store_true')
    parser_statistics.set_defaults(func=statistics)

    parser_statistics_pipeline = subparsers.add_parser(
                        'statistics_pipeline')
    parser_statistics_pipeline.add_argument(
                        '-m', '--methods',
                        type=str,
                        choices=list(ANONYMIZATION_METHODS.keys()),
                        nargs='+',
                        default=list(ANONYMIZATION_METHODS.keys()))
    parser_statistics_pipeline.add_argument(
                        '-s', '--subjects', type=str,
                        choices=src.config.SUBJECTS, nargs='+',
                        default=src.config.SUBJECTS)
    parser_statistics_pipeline.add_argument(
                        '-p', '--pipeline',
                        required=True,
                        type=ProcessingPipeline.from_string,
                        choices=list(ProcessingPipeline))
    parser_statistics_pipeline.add_argument(
                        '-skip_o', '--skip_original',
                        action='store_true')
    group_statistics_pipeline = (parser_statistics_pipeline
                                 .add_mutually_exclusive_group())
    group_statistics_pipeline.add_argument(
                        '-c', '--cost',
                        type=AnonymizationSetting.from_string,
                        choices=(ANONYMIZATION_METHODS['pydeface']
                                 .settings['cost']
                                 .union(
                                   ANONYMIZATION_METHODS['afni_refacer_deface']
                                   .settings['cost'])))
    group_statistics_pipeline.add_argument(
                        '-B', '--bias_correct', action='store_true')
    group_statistics_pipeline.add_argument(
                        '-a', '--afni_mask', action='store_true')
    parser_statistics_pipeline.set_defaults(
                        func=statistics_pipeline)

    parser_cost_statistics_pipeline = subparsers.add_parser(
                        'cost_statistics_pipeline')
    parser_cost_statistics_pipeline.add_argument(
                        '-s', '--subjects', type=str,
                        choices=src.config.SUBJECTS, nargs='+',
                        default=src.config.SUBJECTS)
    parser_cost_statistics_pipeline.add_argument(
                        '-m', '--method',
                        type=str,
                        required=True,
                        choices=[
                            'pydeface',
                            'afni_refacer_deface',
                            'afni_refacer_reface',
                        ])
    parser_cost_statistics_pipeline.add_argument(
                        '-p', '--pipeline',
                        required=True,
                        type=ProcessingPipeline.from_string,
                        choices=list(ProcessingPipeline))
    parser_cost_statistics_pipeline.add_argument(
                        '-skip_o', '--skip_original', action='store_true')
    parser_cost_statistics_pipeline.add_argument(
                        '-skip_c', '--skip_default_cost', action='store_true')
    parser_cost_statistics_pipeline.set_defaults(func=cost_statistics_pipeline)

    return parser.parse_args()


def main():
    if (len(sys.argv) < 2):
        logger.exception('No subcommand given.')
        raise ValueError('No subcommand given.')

    if (sys.argv[1] == 'start' and len(
            os.listdir(src.config.WORKING_DIR)) != 0):
        logger.exception('WORKING_DIR needs to be empty.')
        raise ValueError('WORKING_DIR needs to be empty.')

    initialized = start.load_subjects()
    if (not initialized and sys.argv[1] != 'start'):
        logger.exception(('WORKING_DIR is not initialized. '
                          'Please initialize it with command '
                          'python3 main.py start'))
        raise ValueError(('WORKING_DIR is not initialized. '
                          'Please initialize it with command '
                          'python3 main.py start'))

    args = configure_parsers()
    args.func(args)


if __name__ == '__main__':
    main()
