import logging
from openpyxl import Workbook
import os
import pandas as pd
import shutil
import subprocess

from src.config import STATISTICS_DIR
from src.path_managing import get_bids_root_dir
from src.structures import AnonymizationMethod, \
    AnonymizationSetting, ORIGINAL


def configure_logger(logger):
    ''' Configures logger settings. '''
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

    file_handler = logging.FileHandler('app.log')
    file_handler.setLevel(logging.ERROR)
    file_handler.setFormatter(formatter)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)

    logger.addHandler(file_handler)
    logger.addHandler(stream_handler)


logger = logging.getLogger(__name__)
configure_logger(logger)


def append_duration_excel(subject_name: str, method: AnonymizationMethod,
                          setting: AnonymizationSetting, execution_time,
                          skullstrip_execution_time, skullstrip_name: str,
                          sheet: str) -> pd.DataFrame:
    ''' Appends anonymization execution time to excel sheet. '''
    df = pd.DataFrame()
    df['Subject'] = [subject_name]
    df['Method'] = [str(method)]
    df['Setting'] = [str(setting)]
    df['Execution Time'] = [execution_time]
    if (skullstrip_execution_time is not None):
        df[f'{skullstrip_name} Execution Time'] = [skullstrip_execution_time]

    if (sheet is not None):
        append_data_excel(df, sheet)

    return df


def add_subject_to_participants(subject_name: str, bids_dir_path: str):
    ''' Adds subject to participants.tsv in BIDS directory. '''
    original_bids_root_dir = get_bids_root_dir(ORIGINAL, None)
    original_participants_file_path = os.path.join(original_bids_root_dir,
                                                   'participants.tsv')

    method_participants_file_path = os.path.join(bids_dir_path,
                                                 'participants.tsv')

    tsv_original = pd.read_csv(original_participants_file_path,
                               sep='\t', header=0)
    participant_row = tsv_original.loc[(
                        tsv_original['participant_id'] == subject_name)]

    if (os.path.exists(method_participants_file_path)):
        participant_row.to_csv(method_participants_file_path, mode='a',
                               header=False, index=False, sep='\t')
    else:
        participant_row.to_csv(method_participants_file_path, mode='w',
                               header=True, index=False, sep='\t')


def copy_json_file(subject_name: str, method: AnonymizationMethod,
                   anonymization_settings: AnonymizationSetting) -> None:
    ''' Copies subject json file to BIDS dir of anonymization method. '''
    original_bids_root_dir = get_bids_root_dir(ORIGINAL, None)
    original_subject_json = os.path.join(original_bids_root_dir, subject_name,
                                         'anat', subject_name + '_T1w.json')

    if (os.path.exists(original_subject_json)):
        anonym_bids_root_dir = get_bids_root_dir(method,
                                                 anonymization_settings)
        anonym_subject_json = os.path.join(
                        anonym_bids_root_dir, subject_name, 'anat',
                        subject_name + '_' + method.file_suffix + '_T1w.json')
        shutil.copy2(original_subject_json, anonym_subject_json)


def append_data_excel(dataframe: pd.DataFrame, sheet_name: str) -> None:
    ''' Appends computed data to excel. '''
    sheets_path = os.path.join(STATISTICS_DIR, 'sheets')
    if (not os.path.exists(sheets_path)):
        os.makedirs(sheets_path)

    excel_sheet_path = os.path.join(sheets_path, sheet_name + '.xlsx')

    # create sheet with header if it doesn't exist
    if (not os.path.exists(excel_sheet_path)):
        workbook = Workbook(sheet_name)
        workbook.create_sheet('Data')

        with pd.ExcelWriter(excel_sheet_path, engine='openpyxl') as writer:
            dataframe.to_excel(writer,  sheet_name='Data',
                               header=True, index=True)
            if isinstance(dataframe.keys(), pd.MultiIndex):
                # get rid of unnecessary row
                writer.sheets['Data'].delete_rows(3)

    else:
        with pd.ExcelWriter(excel_sheet_path, engine='openpyxl',
                            if_sheet_exists='overlay', mode='a') as writer:

            # handle index and row
            if isinstance(dataframe.keys(), pd.MultiIndex):
                header_rows = 2
                row = writer.sheets['Data'].max_row - 1
            else:
                header_rows = 1
                row = writer.sheets['Data'].max_row

            dataframe.to_excel(writer, sheet_name='Data', startrow=row,
                               header=False, index=True)
            # index from zero
            writer.sheets['Data'].cell(row + header_rows, 1).value = row - 1


def structure_mri_reface(subject_name: str, method: AnonymizationMethod,
                         method_subject_dir: str, original_subject_file: str):
    ''' Structures output from mri_reface so it is conformed
    to BIDS directiry. '''
    # move refaced subject
    mri_created_folder = os.path.join(method_subject_dir,
                                      (subject_name + '_'
                                       + method.file_suffix + '_T1w.nii'))
    old_defaced = os.path.join(mri_created_folder,
                               subject_name + '_T1w_deFaced.nii')
    new_defaced = os.path.join(method_subject_dir,
                               (subject_name
                                + '_' + method.file_suffix + '_T1w.nii.gz'))
    os.system(f'gzip {old_defaced} -f')
    os.system('mv {} {}'.format(old_defaced + '.gz', new_defaced))

    # delete old folder and .nii original file
    shutil.rmtree(mri_created_folder)
    os.remove(original_subject_file)


def zip_files(directory: str):
    ''' Zips all files in directory. '''
    for filename in os.listdir(directory):
        if filename.endswith('.nii'):
            f = os.path.join(directory, filename)
            os.system(f'gzip {f} -f')


def execute_command(command: str):
    ''' Executes command in OS, raises CalledProcessError if command fails. '''
    try:
        subprocess.run(command, check=True, shell=True,
                       stderr=subprocess.PIPE, encoding='utf-8')
    except subprocess.CalledProcessError as e:
        logger.exception(e.stderr)
        raise
