from nipype.interfaces import fsl
import os
import shutil
from time import perf_counter
from typing import List, Optional

from src.data_manipulation import logger, append_duration_excel, \
    add_subject_to_participants, copy_json_file, structure_mri_reface, \
    execute_command
from src.path_managing import get_bids_root_dir, make_path
from src.start import create_bids_directory
from src.structures import AnonymizationMethod, AnonymizationSetting, ORIGINAL


def execute_anonymization(subjects: List[str],
                          method: AnonymizationMethod,
                          setting: AnonymizationSetting,
                          sheet: str = None) -> None:
    """Executes anonymization of subject images by chosen method with
    chosen settings. Assumes that seetting is for the chosen method correct,
    as inputs to the program are checked.

    Args:
        subjects (List[str]): <SUBJECT_LABEL>s of images which will be
                            anonymized
        method (AnonymizationMethod): tool used for anonymization
        setting (AnonymizationSetting): method setting, defaults to None
        sheet (str, optional): name of excel sheet where duration of
                            anonymization will be written,
                            defaults to None.
    """
    for subject in subjects:

        subject_name = 'sub-' + subject
        logger.info(f'{str(method)} anonymization with {str(setting)} '
                    f'setting for {subject_name}')

        method_bids_root_dir = get_bids_root_dir(method, setting)
        if (not os.path.exists(method_bids_root_dir)):
            create_bids_directory(method, setting)
        original_bids_root_dir = get_bids_root_dir(ORIGINAL)

        original_filename = subject_name + '_T1w.nii.gz'
        anonymized_filename = (subject_name + '_'
                               + method.file_suffix + '_T1w.nii.gz')

        original_subject_file = os.path.join(original_bids_root_dir,
                                             subject_name, 'anat',
                                             original_filename)
        method_subject_dir = os.path.join(method_bids_root_dir,
                                          subject_name, 'anat')
        anonymized_subject_file = os.path.join(method_subject_dir,
                                               anonymized_filename)

        setting_command = get_command_for_setting(method, setting)

        # create path
        if (not os.path.exists(method_subject_dir)):
            os.makedirs(method_subject_dir)

        # is new participant?
        new_participant = False
        if (not os.path.exists(anonymized_subject_file)):
            new_participant = True

        # anonymize original file
        start_skullstrip = None
        execution_time_skullstrip = None
        skullstrip_name = None

        # mri_reface needs .nii file
        if (method.name == 'mri_reface'):
            os.system(f'gunzip {original_subject_file} -k -f')
            original_subject_file = original_subject_file[:-3]
            anonymized_subject_file = anonymized_subject_file[:-3]

        # QUICKSHEAR needs brain mask
        if (method.name == 'quickshear'):
            skullstrip_name = ('BET' if setting == AnonymizationSetting.DEFAULT
                               else '3dSkullStrip')
            output_dir = os.path.join(method_bids_root_dir, 'derivatives')
            brain_subject_file, start_skullstrip, stop_skullstrip = (
                        create_brain_mask(original_subject_file,
                                          output_dir, subject_name, setting))
            command = (method.command.format(original_subject_file,
                                             brain_subject_file,
                                             anonymized_subject_file)
                       + setting_command)
        else:
            command = method.command.format(
                            original_subject_file,
                            anonymized_subject_file) + setting_command

        logger.info('anonymization initialized')
        start = perf_counter()
        execute_command(command)
        stop = perf_counter()

        execution_time = stop - start
        if (start_skullstrip):
            execution_time_skullstrip = stop_skullstrip - start_skullstrip
        logger.info('\n' +
                    append_duration_excel(subject_name, method, setting,
                                          execution_time,
                                          execution_time_skullstrip,
                                          skullstrip_name, sheet).to_string())

        if (method.name == 'mri_reface'):
            # mri_reface creates separate folder which breaks file structure
            structure_mri_reface(subject_name, method, method_subject_dir,
                                 original_subject_file)

        # copy subject json and add subject to participants.tsv
        if (new_participant):
            copy_json_file(subject_name, method, setting)
            if (os.path.exists(os.path.join(original_bids_root_dir,
                                            'participants.tsv'))):
                add_subject_to_participants(subject_name, method_bids_root_dir)


def create_brain_mask(in_file: str, out_dir: str, out_prefix: str,
                      setting: AnonymizationSetting):
    ''' Creates brain mask. '''
    logger.info(f'generating brain mask with {str(setting)} setting')
    if (setting == AnonymizationSetting.DEFAULT):
        out_brain_file_path = os.path.join(
                        out_dir, 'fsl',
                        out_prefix + '_bet_brain_T1w.nii.gz')
        make_path(out_brain_file_path)

        bet = fsl.BET()
        bet.inputs.in_file = in_file
        bet.inputs.frac = 0.5
        bet.inputs.out_file = out_brain_file_path
        start_skullstrip = perf_counter()
        bet.run()
        stop_skullstrip = perf_counter()

    else:
        # nypipe Skullstrip() bug - needs out_file to be BRIK not BRIK.gz
        # which is actually generated so nypipe can not be used
        out_dir = os.path.join(out_dir, 'afni')
        out_brain_file_prefix = os.path.join(
                                            out_dir,
                                            out_prefix + '_3dSkullStrip_brain')

        out_tmp_dir = os.path.join(out_dir, 'tmp_afni_working_dir')
        tmp_out_brain_file_prefix = os.path.join(
                                            out_tmp_dir,
                                            out_prefix + '_3dSkullStrip_brain')
        make_path(tmp_out_brain_file_prefix)

        start_skullstrip = perf_counter()
        command = (f'3dSkullStrip -input {in_file} -prefix '
                   f'{tmp_out_brain_file_prefix} -mask_vol -orig_vol')
        execute_command(command)
        stop_skullstrip = perf_counter()

        # figure out output name (if deoblique dataset,
        # output is tlrc - normalized, not in tlrc space)
        endings = ['+orig.BRIK.gz', '+orig.BRIK',
                   '+tlrc.BRIK.gz', '+tlrc.BRIK']
        for ending in endings:
            if (os.path.exists(tmp_out_brain_file_prefix + ending)):
                out_afni_brain_file = tmp_out_brain_file_prefix + ending
                break

        out_brain_file = out_prefix + '_T1w.nii.gz'
        # 3dAFNItoNIFTI gives error if prefix is not specified
        # 3dAFNItoNIFTI gives error if prefix is the same as one used for
        # input file
        # 3dAFNItoNIFTI prefix doesn't work
        command = (f'3dAFNItoNIFTI {out_afni_brain_file} '
                   f'-prefix {out_brain_file}')
        execute_command(command)

        # since 3dAFNItoNIFTI prefix doesn't work,
        # we need to manually gzip and move generated file
        filename = out_prefix + '_3dSkullStrip_brain.nii'
        os.system(f'gzip {filename} -f')
        # move generated file to the correct output_dir
        out_brain_file_path = out_brain_file_prefix + '_T1w.nii.gz'
        os.system(f'mv {filename}.gz {out_brain_file_path} -f')
        # delete all unnecessary generated files from 3dSkullStrip
        shutil.rmtree(out_tmp_dir)

    return out_brain_file_path, start_skullstrip, stop_skullstrip


def get_command_for_setting(method, setting):
    ''' Generates flags for addition to method's command. '''
    flag = get_flag(method, setting)
    if (flag is not None and setting != AnonymizationSetting.BRAIN_MASK_AFNI):
        if (len(method.settings[flag]) > 1):  # flag has argument
            if (method.name == 'pydeface'):
                return f' --{flag} {str(setting)}'
            return f' -{flag} {str(setting)}'
        return f' -{flag}'
    return ''  # no additional setting


def get_flag(method: AnonymizationMethod,
             setting: AnonymizationSetting) -> Optional[str]:
    ''' Gets flag which will be used for setting
    in anonymization tool command '''
    if (setting == AnonymizationSetting.DEFAULT):
        return None

    for flag, options in method.settings.items():
        if setting in options:
            return flag

    logger.exception((f'Incorrect setting {str(setting)} for {str(method)}.'
                      ' Valid settings include: {a_method.settings}'))
    raise ValueError((f'Incorrect setting {str(setting)} for {str(method)}.'
                      ' Valid settings include: {a_method.settings}'))
