from typing import List, Union
import os

from src.data_manipulation import logger, execute_command
from src.structures import Method, AnonymizationMethod, AnonymizationSetting, \
    MethodDescription, ProcessingPipeline, Space, ORIGINAL
from src.path_managing import get_processing_pipeline_output_dir, \
    get_file_path, get_spm_file_name, get_bids_root_dir, \
    get_processing_pipeline_root_dir


def spm_normalization(subjects: List[str],
                      method: Union[Method, AnonymizationMethod],
                      anonymization_setting: AnonymizationSetting = None
                      ) -> None:
    """Normalizes with SPM anatomical images of subjects generated
    by method with anonymization_setting.

    Args:
        subjects (List[str]): <SUBJECT_LABEL>s of images which will be
                            anonymized
        method (Union[Method, AnonymizationMethod]): tool used for
                            anonymization or original
        anonymization_setting (AnonymizationSetting, optional): method setting,
                            defaults to None if method is original

    Raises:
        ValueError: image file to normalize doesn't exist
    """
    logger.info('SPM normalization initialized')
    for subject in subjects:
        subject_name = 'sub-' + subject
        method_desc = MethodDescription(
                        method, anonymization_setting,
                        ProcessingPipeline.SPM, Space.SUBJECT_SPACE)
        file_path = get_file_path(subject_name, method_desc)
        output_folder_path = get_processing_pipeline_output_dir(subject_name,
                                                                method_desc)

        if (not os.path.exists(file_path)):
            logger.exception(f'file for {subject_name} of {str(method)} '
                             'doesn\'t exist, can not process with spm')
            raise ValueError(f'file for {subject_name} of {str(method)} '
                             'doesn\'t exist, can not process with spm')

        command = ('/opt/spm12/run_spm12.sh /opt/mcr/v97 script '
                   f'./src/spm_norm.m {file_path} {output_folder_path}')
        execute_command(command)


def spm_normalization_from_original_warp(
                            subjects: List[str],
                            method: AnonymizationMethod,
                            anonymization_setting: AnonymizationSetting = None
                            ) -> None:
    """Normalizes with SPM anatomical images of subjects generated by method
    with anonymization_setting, uses warp calculated from original image.

    Args:
        subjects (List[str]): <SUBJECT_LABEL>s of images which will be
                            anonymized
        method (AnonymizationMethod): tool used for anonymization or original
        anonymization_setting (AnonymizationSetting, optional): method setting,
                            defaults to None if method is original
    """
    logger.info('SPM normalization initialized')
    for subject in subjects:
        subject_name = 'sub-' + subject
        method_desc = MethodDescription(
                        method, anonymization_setting,
                        ProcessingPipeline.SPM, Space.SUBJECT_SPACE)
        original_desc = MethodDescription(
                        ORIGINAL, None,
                        ProcessingPipeline.SPM, Space.SUBJECT_SPACE)

        file_path = get_file_path(subject_name, method_desc)
        output_folder_path = get_processing_pipeline_output_dir(subject_name,
                                                                method_desc)
        original_output_folder_path = get_processing_pipeline_output_dir(
                                                subject_name, original_desc)
        original_y_path = os.path.join(original_output_folder_path,
                                       'y_' + get_spm_file_name(
                                           subject_name, original_desc))

        if (not os.path.exists(original_y_path)):
            original_file_path = get_file_path(subject_name, original_desc)

            command = ('/opt/spm12/run_spm12.sh /opt/mcr/v97 script '
                       f'./src/spm_norm.m {original_file_path} '
                       f'{original_output_folder_path}')
            execute_command(command)

        command = ('/opt/spm12/run_spm12.sh /opt/mcr/v97 script '
                   f'./src/spm_norm_from_warp.m {original_y_path} '
                   f'{file_path} {output_folder_path}')
        execute_command(command)


def fmriprep_anat_pipeline(subjects: List[str],
                           method: Union[Method, AnonymizationMethod],
                           anonymization_setting: AnonymizationSetting = None
                           ) -> None:
    """Executes fMRIPrep anatomical pipeline of subjects generated
    by method with anonymization_setting.

    Args:
        subjects (List[str]): <SUBJECT_LABEL>s of images which will be
                            anonymized
        method (Union[Method, AnonymizationMethod]): tool used for
                            anonymization or original
        anonymization_setting (AnonymizationSetting, optional): method setting,
                            defaults to None if method is original

    Raises:
        ValueError: image file to normalize doesn't exist
    """
    method_desc = MethodDescription(method, anonymization_setting,
                                    ProcessingPipeline.FMRIPREP,
                                    Space.SUBJECT_SPACE)
    method_bids_root_dir = get_bids_root_dir(method, anonymization_setting)
    method_derivatives_dir = get_processing_pipeline_root_dir(method_desc)

    if (not os.path.exists(method_bids_root_dir)):
        logger.exception(f'BIDS directory for {str(method)} doesn\'t exist, '
                         'can not process with fmriprep')
        raise ValueError(f'BIDS directory for {str(method)} doesn\'t exist, '
                         'can not process with fmriprep')

    subjects_input = ' '.join(subjects)
    command = (f'fmriprep {method_bids_root_dir} {method_derivatives_dir} '
               'participant '
               f'--participant-label {subjects_input} '
               '--skip-bids-validation '
               '--md-only-boilerplate '
               '--fs-license-file /opt/freesurfer/license.txt '
               '--fs-no-reconall '
               '--output-spaces MNI152NLin2009cAsym:res-1 '
               '--stop-on-first-crash '
               '--anat-only '
               '-w $HOME')

    logger.info('FMRIPREP pipeline initialized')
    execute_command(command)
