import numpy as np
import nibabel as nib
from nilearn.image import math_img

from src.path_managing import get_tissue_file_path, \
    get_fmriprep_brain_mask_path, get_file_path
from src.structures import MethodDescription, TissueType, Space, \
    ProcessingPipeline, ORIGINAL


def to_Nifti1Image(array: np.ndarray, source: nib.nifti1.Nifti1Image
                   ) -> nib.nifti1.Nifti1Image:
    ''' Creates NiftiiImage from np.ndarray '''
    return nib.Nifti1Image(array, source.affine)


def create_mask(image: nib.nifti1.Nifti1Image) -> nib.nifti1.Nifti1Image:
    ''' Creates binary mask from NiftiiImage '''
    mask = math_img('img > 0.1', img=image)
    return mask


def mask_image(image: nib.nifti1.Nifti1Image, mask: nib.nifti1.Nifti1Image
               ) -> nib.nifti1.Nifti1Image:
    '''Masks image by mask '''
    mask_array = mask.get_fdata()
    image_array = image.get_fdata()

    new_image_array = np.zeros(image.shape)
    new_image_array[mask_array == 1] = image_array[mask_array == 1]

    image_masked = nib.Nifti1Image(new_image_array, image.affine)
    return image_masked


def create_tissue_mask(subject_name: str, method_desc: MethodDescription,
                       tissue_type: TissueType) -> nib.nifti1.Nifti1Image:
    ''' Creates mask for tissue. '''
    tissue_path = get_tissue_file_path(subject_name, method_desc, tissue_type)
    tissue = nib.load(tissue_path)
    return create_mask(tissue)


def create_brain_mask(subject_name, method_desc) -> nib.Nifti1Image:
    ''' Creates brain mask from generated segmented tissues '''
    WM_mask = create_tissue_mask(subject_name, method_desc, TissueType.WM)
    GM_mask = create_tissue_mask(subject_name, method_desc, TissueType.GM)
    CSF_mask = create_tissue_mask(subject_name, method_desc, TissueType.CSF)

    brain_mask_array = (np.array(WM_mask.dataobj)
                        | np.array(GM_mask.dataobj)
                        | np.array(CSF_mask.dataobj))
    brain_mask = to_Nifti1Image(brain_mask_array, WM_mask)
    return brain_mask


def make_masked_image(subject_name: str, method_desc: MethodDescription
                      ) -> nib.nifti1.Nifti1Image:
    ''' Mask subject image with correct mask. '''
    if (method_desc.space == Space.SUBJECT_SPACE):

        # subject_space, make or load brain_mask from original segmentation
        original_desc = MethodDescription(ORIGINAL, None,
                                          method_desc.processing_pipeline,
                                          method_desc.space)
        if (method_desc.processing_pipeline == ProcessingPipeline.FMRIPREP):
            brain_mask_path = get_fmriprep_brain_mask_path(subject_name,
                                                           original_desc)
            brain_mask = nib.load(brain_mask_path)
        else:
            brain_mask = create_brain_mask(subject_name, original_desc)

    else:
        if (method_desc.processing_pipeline == ProcessingPipeline.FMRIPREP):
            brain_mask = nib.load(
                'data/MNI152NLin2009cAsym_res-01_desc-brain_mask.nii.gz')
        else:
            brain_mask = nib.load('data/rmask_ICV.nii.gz')

    image_path = get_file_path(subject_name, method_desc)
    image = nib.load(image_path)

    return mask_image(image, brain_mask)
