import os
from typing import Optional

from src.config import STATISTICS_DIR, WORKING_DIR
from src.structures import AnonymizationMethod, MethodDescription, \
    StatisticsType, ORIGINAL, Mask, GeneratedFile, Space, \
    AnonymizationSetting, ProcessingPipeline, TissueType, Method


def put_underscore(method: AnonymizationMethod) -> str:
    ''' Returns underscore if method adds file_suffix. '''
    return '' if method.file_suffix == '' else '_'


def get_statistics_output_folder_path(subject_name: str,
                                      method_desc: MethodDescription,
                                      statistics_type: StatisticsType) -> str:
    ''' Returns statistics output folder path. '''
    if (method_desc.method.name == ORIGINAL.name):
        # for original we won't create files in separate subfolders
        statistics_folder = ''
    else:
        statistics_folder = str(statistics_type)

    return os.path.join(
                STATISTICS_DIR, subject_name, str(method_desc.method),
                str(method_desc.anonymization_setting or ''),
                str(method_desc.processing_pipeline), statistics_folder)


def get_generated_file_path(subject_name: str, method_desc: MethodDescription,
                            mask: Mask, file_suffix: GeneratedFile,
                            statistics_type: StatisticsType) -> str:
    ''' Returns path of generated file. '''
    output_folder = get_statistics_output_folder_path(
                        subject_name, method_desc, statistics_type)

    space = ('' if method_desc.space == Space.SUBJECT_SPACE
             else '_space-MNI152_res-1')
    # None e.g for warped_facemask
    mask_name = '' if mask is None else str(mask)
    underscore = '' if mask == '' else '_'
    description = '_desc-' + mask_name + underscore + str(file_suffix)
    name = (subject_name + put_underscore(method_desc.method)
            + method_desc.method.file_suffix)
    modality = '_T1w' if file_suffix == GeneratedFile.MASKED else ''

    return os.path.join(output_folder,
                        name + space + description + modality + '.nii.gz')


def make_generated_file_path(subject_name: str, method_desc: MethodDescription,
                             mask_type: Mask, file_suffix: GeneratedFile,
                             statistics_type: StatisticsType) -> str:
    ''' Creates path for generated file. '''
    path = get_generated_file_path(subject_name, method_desc, mask_type.name,
                                   file_suffix, statistics_type)
    make_path(path)
    return path


def make_path(image_path):
    ''' Creates path to file directory. '''
    directory = os.path.dirname(image_path)
    if (not os.path.exists(directory)):
        os.makedirs(directory)


def generate_sheet_root_name(method: AnonymizationMethod,
                             setting: AnonymizationSetting):
    ''' Generates root name for excel sheet. '''
    prefix = ''
    if (method is not None and setting != AnonymizationSetting.DEFAULT):
        if ('COST' in setting.name):
            prefix += method.name + '_cost'
        else:
            prefix += method.name + '_' + str(setting)
    return prefix


def get_sheet_name(statistics_type: StatisticsType,
                   processing_pipeline: ProcessingPipeline,
                   method: AnonymizationMethod = None,
                   setting: AnonymizationSetting = None):
    ''' Returns excel sheet name. '''
    sheet_root_name = generate_sheet_root_name(method, setting)
    underscore = '_' if sheet_root_name else ''
    return (str(processing_pipeline) + underscore
            + sheet_root_name + '_' + str(statistics_type))


def get_sheet_path(statistics_type: StatisticsType,
                   processing_pipeline: ProcessingPipeline,
                   method: AnonymizationMethod = None,
                   setting: AnonymizationSetting = None):
    ''' Returns excel sheet path. '''
    return os.path.join(STATISTICS_DIR, 'sheets',
                        get_sheet_name(statistics_type, processing_pipeline,
                                        method, setting) + '.xlsx')


def get_spm_file_name(subject_name: str,
                      method_desc: MethodDescription) -> str:
    ''' Returns file name from SPM. '''
    return (method_desc.space.spm_file_prefix + subject_name
            + put_underscore(method_desc.method)
            + method_desc.method.file_suffix + '_T1w.nii.gz')


def get_spm_tissue_file_name(subject_name: str,
                             method_desc: MethodDescription,
                             tissue_type: TissueType):
    ''' Returns tissue file name from SPM. '''
    return (tissue_type.spm_label + subject_name +
            put_underscore(method_desc.method)
            + method_desc.method.file_suffix + '_T1w.nii.gz')


def get_fmriprep_file_name(subject_name: str,
                           method_desc: MethodDescription) -> str:
    ''' Returns file name from fMRIPrep. '''
    if (method_desc.space == Space.SUBJECT_SPACE):
        return subject_name + '_desc-preproc_T1w.nii.gz'
    else:
        return (subject_name
                + '_space-MNI152NLin2009cAsym_res-1_desc-preproc_T1w.nii.gz')


def get_fmriprep_tissue_file_name(subject_name: str,
                                  method_desc: MethodDescription,
                                  tissue_type: TissueType) -> str:
    ''' Returns tissue file name from fMRIPrep. '''
    if (method_desc.space == Space.SUBJECT_SPACE):
        return subject_name + '_label-' + tissue_type.name + '_probseg.nii.gz'
    else:
        return (subject_name + '_space-MNI152NLin2009cAsym_res-1_label-'
                + tissue_type.name + '_probseg.nii.gz')


def get_fmriprep_brain_mask_path(subject_name: str,
                                 method_desc: MethodDescription) -> str:
    ''' Return path of brain mask generated by fMRIPrep. '''
    root_folder = get_processing_pipeline_root_dir(method_desc)
    if (method_desc.space == Space.SUBJECT_SPACE):
        file_suffix = '_desc-brain_mask.nii.gz'
    else:
        file_suffix = '_space-MNI152NLin2009cAsym_res-1_desc-brain_mask.nii.gz'

    return os.path.join(root_folder, subject_name, 'anat',
                        subject_name + file_suffix)


def get_bids_root_dir(
                method: Method,
                anonymization_setting: Optional[AnonymizationSetting] = None
                ) -> str:
    ''' Returns BIDS root dir path for method. '''
    return os.path.join(WORKING_DIR, str(method),
                        str(anonymization_setting or ''))


def get_root_dir(method_desc: MethodDescription) -> str:
    ''' Returns BIDS root dir path for file described by method_desc. '''
    if (method_desc.processing_pipeline == ProcessingPipeline.SPM
            and method_desc.space == Space.SUBJECT_SPACE):
        return get_bids_root_dir(method_desc.method,
                                 method_desc.anonymization_setting)
    return get_processing_pipeline_root_dir(method_desc)


def get_processing_pipeline_root_dir(method_desc: MethodDescription) -> str:
    ''' Returns root dir for processing pipeline in derivatives. '''
    return os.path.join(WORKING_DIR, str(method_desc.method),
                        str(method_desc.anonymization_setting or ''),
                        'derivatives', str(method_desc.processing_pipeline))


def get_processing_pipeline_output_dir(subject_name: str,
                                       method_desc: MethodDescription) -> str:
    ''' Returns output directory of processing pipeline
    for concrete subject and method. '''
    pipeline_dir = get_processing_pipeline_root_dir(method_desc)
    return os.path.join(pipeline_dir, subject_name, 'anat')


def get_tissue_file_path(subject_name: str,
                         method_desc: MethodDescription,
                         tissue_type: TissueType) -> str:
    ''' Returns tissue_type file path of file described
    by method_desc of subject_name. '''
    processing_pipeline_root = get_processing_pipeline_root_dir(method_desc)
    if (method_desc.processing_pipeline == ProcessingPipeline.FMRIPREP):
        file_name = get_fmriprep_tissue_file_name(subject_name, method_desc,
                                                  tissue_type)
    else:
        file_name = get_spm_tissue_file_name(subject_name, method_desc,
                                             tissue_type)
    return os.path.join(processing_pipeline_root, subject_name, 'anat',
                        file_name)


def get_file_path(subject_name: str, method_desc: MethodDescription) -> str:
    ''' Returns file path of file described by method_desc of subject_name. '''
    root_dir = get_root_dir(method_desc)
    if (method_desc.processing_pipeline == ProcessingPipeline.FMRIPREP):
        file_name = get_fmriprep_file_name(subject_name, method_desc)
    else:
        file_name = get_spm_file_name(subject_name, method_desc)

    return os.path.join(root_dir, subject_name, 'anat', file_name)
