from typing import List

from src.structures import AnonymizationMethod, AnonymizationSetting, \
    ProcessingPipeline, ORIGINAL, ANONYMIZATION_METHODS
from src.execute_anonymization import execute_anonymization
from src.execute_pipeline_processing import spm_normalization, \
    spm_normalization_from_original_warp, fmriprep_anat_pipeline
from src.path_managing import generate_sheet_root_name
from src.statistics import execute_statistics, execute_cost_statistics


def statistics_pipeline(subjects: List[str],
                        methods: List[AnonymizationMethod],
                        a_setting: AnonymizationSetting,
                        processing_pipeline: ProcessingPipeline,
                        skip_orig_processing: bool = False) -> None:
    ''' Combines anonymization, data processing with
    processing_pipeline and statistics into one call. '''
    execute_pipeline = (
        spm_normalization if processing_pipeline == ProcessingPipeline.SPM
        else fmriprep_anat_pipeline)

    # original
    if (not skip_orig_processing):
        execute_pipeline(subjects, ORIGINAL, None)

    # anonymize and normalize images
    sheet_prefix = generate_sheet_root_name(methods[0], a_setting)
    if (sheet_prefix == ''):
        sheet_prefix = 'default'
    sheet_name = sheet_prefix + '_anonymization_execution_time'
    for method in methods:
        execute_anonymization(subjects, method, a_setting, sheet_name)
        execute_pipeline(subjects, method, a_setting)

    if (processing_pipeline == ProcessingPipeline.SPM):
        for method in methods:
            spm_normalization_from_original_warp(subjects, method, a_setting)

    # create statistics
    execute_statistics(subjects, methods, a_setting, processing_pipeline)


def cost_statistics_pipeline(subjects: List[str],
                             method: AnonymizationMethod,
                             processing_pipeline: ProcessingPipeline,
                             skip_default_cost: bool = False,
                             skip_orig_processing: bool = False) -> None:
    ''' Combines anonymization, data processing with processing_pipeline and
    statistics of all available cost functions into one call. '''
    execute_pipeline = (
        spm_normalization if processing_pipeline == ProcessingPipeline.SPM
        else fmriprep_anat_pipeline)

    if (not skip_orig_processing):
        execute_pipeline(subjects, ORIGINAL, None)

    sheet_name = method.name + '_cost_anonymization_execution_time'
    cost_options = method.settings['cost']
    if (skip_default_cost):
        if (method == ANONYMIZATION_METHODS['pydeface']):
            cost_options.remove(AnonymizationSetting.COST_MUTUALINFO)
        else:
            cost_options.remove(AnonymizationSetting.COST_LOCAL_PCOR_ABS)

    # anonymize and process images
    for cost_option in cost_options:
        execute_anonymization(subjects, method, cost_option, sheet_name)
        execute_pipeline(subjects, method, cost_option)

    if (processing_pipeline == ProcessingPipeline.SPM):
        for cost_option in cost_options:
            spm_normalization_from_original_warp(subjects, method, cost_option)

    # create statistics
    execute_cost_statistics(subjects, method, processing_pipeline,
                            skip_default_cost)
