from bids import BIDSLayout
from bids.exceptions import BIDSValidationError
import os
import shutil

from src.structures import ORIGINAL, AnonymizationMethod, AnonymizationSetting
from src.data_manipulation import logger, zip_files
import src.config
from src.config import DATA_PATH, WORKING_DIR
from src.path_managing import get_bids_root_dir


def init_working_dir():
    ''' Prepares working directory.
        Validates structure of data in DATA_PATH
        and copies images to working directory.
    '''
    # new working directory needs to be empty
    if (len(os.listdir(WORKING_DIR)) != 0):
        logger.exception(f'WORKING DIRECTORY {WORKING_DIR} is not empty.')
        raise ValueError(f'WORKING DIRECTORY {WORKING_DIR} is not empty.')

    validate_structure()
    copy_bids_original()

    os.mkdir(os.path.join(WORKING_DIR, 'statistics'))


def load_subjects() -> bool:
    ''' Loads subjects to variable from original BIDS directory '''
    original_bids_path = get_bids_root_dir(ORIGINAL)
    if (not os.path.exists(original_bids_path)):
        return False
    try:
        layout = BIDSLayout(original_bids_path)
    except (ValueError, BIDSValidationError) as e:
        logger.exception(e)
        raise
    src.config.SUBJECTS = layout.get_subjects()
    if (not src.config.SUBJECTS):
        logger.exception(f'No subjects found in {original_bids_path}')
    src.config.SUBJECTS.sort()
    return True


def validate_structure():
    ''' Validates structure of data located in DATA_PATH according to BIDS
    standard. Sessions and runs are not supported.'''
    try:
        layout = BIDSLayout(DATA_PATH)
    except (ValueError, BIDSValidationError) as e:
        # invalid BIDS
        logger.exception(e)
        raise

    src.config.SUBJECTS = layout.get_subjects()
    if (not src.config.SUBJECTS):
        logger.exception(f'No subjects found in {DATA_PATH}')
        raise ValueError(f'No subjects found in {DATA_PATH}')
    src.config.SUBJECTS.sort()

    logger.info('RUNNING structure validation - all suject anat files exist')

    for subject in src.config.SUBJECTS:
        subject_name = 'sub-' + subject

        anat_file = os.path.join(DATA_PATH, subject_name, 'anat',
                                 subject_name + '_T1w')
        if (not os.path.exists(anat_file + '.nii') and
                not os.path.exists(anat_file + '.nii.gz')):
            logger.info(
                'Invalid data structure. No runs and sessions are supported.')
            logger.exception(
                f'Invalid data structure. File {anat_file} doesn\'t exist.')
            raise ValueError(
                f'Invalid data structure. File {anat_file} doesn\'t exist.')

    logger.info('BIDS structure : SUCCESS')


def copy_bids_original():
    ''' Copies original anat T1 data with JSON from DATA_PATH to
    WORKING_DIR and zips it if necessary '''
    new_original_root_dir = get_bids_root_dir(ORIGINAL, None)
    logger.info(f'Copying data to {WORKING_DIR}...')

    try:
        os.mkdir(new_original_root_dir)
    except OSError as e:
        logger.exception(e)
        raise

    for subject in src.config.SUBJECTS:
        subject_name = 'sub-' + subject
        anat_file = os.path.join(DATA_PATH, subject_name, 'anat',
                                 subject_name + '_T1w')
        copied_anat_file = os.path.join(new_original_root_dir, subject_name,
                                        'anat', subject_name + '_T1w')
        os.makedirs(os.path.dirname(copied_anat_file))

        if (os.path.exists(anat_file + '.nii.gz')):
            shutil.copy2(anat_file + '.nii.gz', copied_anat_file + '.nii.gz')
        else:
            shutil.copy2(anat_file + '.nii', copied_anat_file + '.nii')

        original_subject_json = os.path.join(
                                DATA_PATH, subject_name,
                                'anat', subject_name + '_T1w.json')
        if (os.path.exists(original_subject_json)):
            copied_subject_json = os.path.join(
                                new_original_root_dir, subject_name,
                                'anat', subject_name + '_T1w.json')
            shutil.copy2(original_subject_json, copied_subject_json)

    for root, dirs, _ in os.walk(new_original_root_dir):
        for name in dirs:
            dir = os.path.join(root, name)
            zip_files(dir)

    for file_name in os.listdir(DATA_PATH):
        source_file_path = os.path.join(DATA_PATH, file_name)
        if (os.path.isfile(source_file_path)):
            destination_file_path = os.path.join(
                        new_original_root_dir, file_name)
            shutil.copy2(source_file_path, destination_file_path)

    os.mkdir(os.path.join(new_original_root_dir, 'derivatives'))


def create_bids_directory(method: AnonymizationMethod,
                          setting: AnonymizationSetting):
    ''' Creates BIDS directory for anonymization method. '''
    logger.info(f'Creating {method} with {setting} settings BIDS directory')

    bids_method_dir = get_bids_root_dir(method, setting)
    assert(not os.path.exists(bids_method_dir))

    bids_original_dir = get_bids_root_dir(ORIGINAL, None)
    if (not os.path.exists(bids_original_dir)):
        logger.info('Original BIDS directory doesn\'t exist')
        copy_bids_original()

    os.makedirs(bids_method_dir)
    participants_filename = 'participants.tsv'

    # copy original BIDS description files
    for file_name in os.listdir(bids_original_dir):

        # participants will be added to participants.tsv when created,
        # since user can anonymize only some of them
        if (file_name == participants_filename):
            continue

        source_file_path = os.path.join(bids_original_dir, file_name)
        if (os.path.isfile(source_file_path)):
            destination_file_path = os.path.join(bids_method_dir, file_name)
            shutil.copy2(source_file_path, destination_file_path)
            logger.info('copied ' + file_name)

    os.mkdir(os.path.join(bids_method_dir, 'derivatives'))
    logger.info(str(method) + ' BIDS directory successfully created')
