import traceback
from typing import List, Union

import numpy as np
import nibabel as nib
import pandas as pd
import nipype.algorithms.metrics as metrics
from skimage import metrics as skimage_metrics
from nilearn.image import math_img
import os

from src.data_manipulation import logger, append_data_excel
import src.masking as masking
from src.path_managing import get_tissue_file_path, \
    make_generated_file_path, get_generated_file_path, \
    get_file_path, get_sheet_name, get_sheet_path
from src.structures import MethodDescription, TissueType, \
    StatisticsType, ORIGINAL, GeneratedFile, Mask, Space, \
    AnonymizationMethod, AnonymizationSetting, ProcessingPipeline


def compute_volume(tissue_image: nib.Nifti1Image) -> float:
    ''' Returns volume of tissue. '''
    tissue_bin = masking.create_mask(tissue_image)
    tissue_array = tissue_bin.get_fdata()
    vols = np.sum(tissue_array)

    return vols


def compute_mae(original_image: Union[str, nib.Nifti1Image],
                anonymized_image: Union[str, nib.Nifti1Image]) -> float:
    ''' Returns MAE value of comparison between original and
    anonymized image. '''
    difference_img = math_img('np.abs(img1 - img2)',
                              img1=original_image, img2=anonymized_image)
    diff_array = difference_img.get_fdata()
    n = diff_array.size
    diff_voxels = np.sum(diff_array)

    mae = diff_voxels / n
    return mae


def compute_mse(original_image: np.ndarray,
                anonymized_image: np.ndarray) -> float:
    ''' Returns MSE value of comparison between original and
    anonymized image. '''
    return skimage_metrics.mean_squared_error(original_image, anonymized_image)


def compute_nmi(original_image: np.ndarray,
                anonymized_image: np.ndarray) -> float:
    ''' Returns NMI value of comparison between original and
    anonymized image. '''
    return skimage_metrics.normalized_mutual_information(original_image,
                                                         anonymized_image)


def compute_structural_similiarity(original_image: np.ndarray,
                                   anonymized_image: np.ndarray) -> float:
    ''' Returns SSIM value of comparison between original and
    anonymized image. '''
    return skimage_metrics.structural_similarity(original_image,
                                                 anonymized_image, full=True)


def compute_dice(original_image_path: str, anonymized_image_path: str,
                 out_file_path: str) -> float:
    """Computes Dice metric.

    Args:
        original_image_path (str): path to image for comparison
        anonymized_image_path (str): path to image for comparison
        out_file_path (str): path for overlap image
        type_name (str): name of tissue being compared

    Returns:
        pd.DataFrame: Dice saved in dataframe
    """
    overlap = metrics.Overlap()
    overlap.inputs.volume1 = original_image_path
    overlap.inputs.volume2 = anonymized_image_path
    overlap.inputs.out_file = out_file_path
    res = overlap.run()
    outputs = res.outputs

    return outputs.dice


def compute_overlap(original_image: Union[str, nib.Nifti1Image],
                    anonymized_image: Union[str, nib.Nifti1Image],
                    out_file: str = None) -> nib.Nifti1Image:
    """Computes overlap difference.

    Args:
        original_image (Union[str, nib.Nifti1Image]): NIFTI image or path to
                                                      one for comparison
        anonymized_image (Union[str, nib.Nifti1Image]): NIFTI image or path to
                                                        one for comparison
        out_file (str, optional): path for overlap image. Defaults to None.

    Returns:
        nib.Nifti1Image: overlap image
    """
    overlap_img = math_img('img1 != img2',
                           img1=original_image, img2=anonymized_image)
    if (out_file):
        overlap_img.to_filename(out_file)
    return overlap_img


def create_statistics(original_image: nib.Nifti1Image,
                      anonym_image: nib.Nifti1Image,
                      ssim_diff_path: str,
                      type_name: str = None) -> pd.DataFrame:
    """Creates statistics for one pair of Niftii images which are being compared.
    Computes MAE, MSE, NMI, SSIM. Saves SSIM diff_image.

    Args:
        original_image (nib.Nifti1Image): original image being compared
        anonym_image (nib.Nifti1Image): anonymized image being compared
        ssim_diff_path (str): path where SSIM difference image will be saved
        type_name (str, optional): name of tissue, if tissues are being
                                   compared. Defaults to None.

    Returns:
        pd.DataFrame: results in dataframe
    """
    original_image_data = original_image.get_fdata()
    anonym_image_data = anonym_image.get_fdata()

    mae = compute_mae(original_image, anonym_image)
    mse = compute_mse(original_image_data, anonym_image_data)
    nmi = compute_nmi(original_image_data, anonym_image_data)

    ssim, ssim_image_array = compute_structural_similiarity(
                                original_image_data, anonym_image_data)
    ssim_image = masking.to_Nifti1Image(ssim_image_array, original_image)
    ssim_image.to_filename(ssim_diff_path)

    data = [mae, mse, nmi, ssim]
    metrics = ['MAE', 'MSE', 'NMI', 'SSIM']

    # multiindex dataframe - metrics for WM | GM | CSF
    if (type_name):
        cols = pd.MultiIndex.from_product([[type_name], metrics])
    else:
        cols = metrics

    df = pd.DataFrame([data], columns=cols)

    return df


def subject_segmented_tissues_statistics(subject_name: str,
                                         method_desc: MethodDescription,
                                         statistics_type: StatisticsType,
                                         sheet: str = None) -> pd.DataFrame:
    """Creates statistics for pairs of segmented tissues of subject.

    Args:
        subject_name (str): name of subject whose generated tissues will be
                            compared, sub-<SUBJECT_LABEL>
        method_desc (MethodDescription): description of how NIFTII image was
                            processed, used to locate relevant files
        statistics_type (StatisticsType): type of statistic being computed,
                            used to save files in relevant folders and write
                            results to concrete excel sheet
        sheet (str, optional): sheet name, where results will be saved,
                            if None, results won't be saved. Defaults to None.

    Returns:
        pd.DataFrame: results in dataframe
    """
    assert statistics_type == StatisticsType.SEGMENTED_TISSUES
    data = [subject_name, str(method_desc.method),
            method_desc.anonymization_setting.setting_name]
    cols = pd.MultiIndex.from_product([['Description'],
                                       ['Subject', 'Method', 'Setting']])
    df = pd.DataFrame([data], columns=cols)

    original_desc = MethodDescription(ORIGINAL, None,
                                      method_desc.processing_pipeline,
                                      method_desc.space)

    for tissue in TissueType:
        original_tissue_path = get_tissue_file_path(subject_name,
                                                    original_desc, tissue)
        anonym_tissue_path = get_tissue_file_path(subject_name,
                                                  method_desc, tissue)

        original_tissue = nib.load(original_tissue_path)
        anonym_tissue = nib.load(anonym_tissue_path)

        ssim_diff_path = make_generated_file_path(
                            subject_name, method_desc, Mask[tissue.name],
                            GeneratedFile.SSIM_DIFF, statistics_type)
        overlap_diff_path = make_generated_file_path(
                            subject_name, method_desc, Mask[tissue.name],
                            GeneratedFile.OVERLAP_DIFF, statistics_type)

        df_addition = create_statistics(original_tissue, anonym_tissue,
                                        ssim_diff_path, tissue.name)
        df = pd.concat([df, df_addition], axis=1)

        df[tissue.name, 'DICE'] = compute_dice(original_tissue_path,
                                               anonym_tissue_path,
                                               overlap_diff_path)

        vols = compute_volume(anonym_tissue)
        df[tissue.name, 'Voxels'] = [vols]

    if (sheet is not None):
        append_data_excel(df, sheet)

    return df


def subject_statistics(subject_name: str, method_desc: MethodDescription,
                       statistics_type: StatisticsType,
                       sheet: str = None) -> pd.DataFrame:
    """Creates statistics for pairs of subject anatomical NIFTII images
    - either in subject_space or in MNI152_space.

    Args:
        subject_name (str): name of subject whose anat NIFTII images will be
                            compared, sub-<SUBJECT_LABEL>
        method_desc (MethodDescription): description of how NIFTII image was
                            processed, used to locate relevant files
        statistics_type (StatisticsType): type of statistic being computed,
                            used to save files in relevant folders and write
                            results to concrete excel sheet
        sheet (str, optional): sheet name, where results will be saved,
                            if None, results won't be saved. Defaults to None.

    Returns:
        pd.DataFrame: results in dataframe
    """
    df = pd.DataFrame()
    df['Subject'] = [subject_name]
    df['Method'] = [str(method_desc.method)]
    df['Setting'] = [method_desc.anonymization_setting.setting_name]

    normalization_type_original = method_desc.space
    if (method_desc.space == Space.MNI152_SPACE_ORIG_WARP):
        normalization_type_original = Space.MNI152_SPACE
    original_desc = MethodDescription(ORIGINAL, None,
                                      method_desc.processing_pipeline,
                                      normalization_type_original)

    original_masked = get_masked_image(subject_name, original_desc,
                                       statistics_type, False)
    anonym_masked = get_masked_image(subject_name, method_desc,
                                     statistics_type)
    ssim_diff_path = make_generated_file_path(subject_name, method_desc,
                                              Mask.BRAIN,
                                              GeneratedFile.SSIM_DIFF,
                                              statistics_type)

    df_addition = create_statistics(original_masked, anonym_masked,
                                    ssim_diff_path)
    df = df.join(df_addition)

    if (sheet is not None):
        append_data_excel(df, sheet)

    return df


def get_masked_image(subject_name: str, method_desc: MethodDescription,
                     statistics_type: StatisticsType,
                     force: bool = True) -> nib.Nifti1Image:
    """Loads already generated masked image (for original) or creates new one.

    Args:
        subject_name (str): name of subject whose anat NIFTII images will be
                            compared, sub-<SUBJECT_LABEL>
        method_desc (MethodDescription): description of how NIFTII image was
                            processed, used to locate relevant files
        statistics_type (StatisticsType): type of statistic being computed,
                            used to locate relevant files
        force (bool, optional): even if image already exists, create new one.
                            Defaults to True.

    Returns:
        nib.Nifti1Image: masked NIfTI image
    """
    file_suffix = GeneratedFile.MASKED
    file_path = get_generated_file_path(subject_name, method_desc,
                                        Mask.BRAIN, file_suffix,
                                        statistics_type)
    if ((not os.path.exists(file_path)) or force):
        # create file
        file = masking.make_masked_image(subject_name, method_desc)
        path = make_generated_file_path(subject_name, method_desc, Mask.BRAIN,
                                        GeneratedFile.MASKED, statistics_type)
        file.to_filename(path)
    else:
        file = nib.load(file_path)

    return file


def execute_statistics(subjects: List[str], methods: List[AnonymizationMethod],
                       setting: AnonymizationSetting,
                       processing_pipeline: ProcessingPipeline) -> None:
    """Calculates statistics for setting in anonymization tools.

    Args:
        subjects (List[str]): subject names whose statistics will be
                            calculated, [sub-<SUBJECT_LABEL>]
        methods (List[AnonymizationMethod]): anonymization methods for whose
                            generated files will be statistics calculated
        processing_pipeline (ProcessingPipeline): pipeline whose results
                            can be altered by anonymization of files
    """
    for subject in subjects:
        subject_name = 'sub-' + subject

        logger.info(subject_name + ':')

        for method in methods:
            logger.info(method.name + ':')

            method_desc = MethodDescription(method, setting,
                                            processing_pipeline,
                                            Space.SUBJECT_SPACE)
            method_desc_mni = MethodDescription(method, setting,
                                                processing_pipeline,
                                                Space.MNI152_SPACE)
            method_desc_mni_original_y = MethodDescription(
                                                method, setting,
                                                ProcessingPipeline.SPM,
                                                Space.MNI152_SPACE_ORIG_WARP)

            stats = {StatisticsType.MNI152_SPACE: method_desc_mni,
                     StatisticsType.SUBJECT_SPACE: method_desc}
            mni_original_y_path = get_file_path(subject_name,
                                                method_desc_mni_original_y)

            if (processing_pipeline == ProcessingPipeline.SPM and
                    os.path.exists(mni_original_y_path)):
                stats[StatisticsType.MNI152_SPACE_ORIG_WARP] = (
                                            method_desc_mni_original_y)

            for statistics_type, m_desc in stats.items():
                sheet = get_sheet_name(statistics_type, processing_pipeline,
                                       method, setting)
                logger.info(
                    f'{subject_name} {str(setting)} {str(statistics_type)}')

                try:
                    subject_statistics(subject_name, m_desc,
                                       statistics_type, sheet)
                except:
                    logger.exception((
                        f'Error while calculating {str(statistics_type)} '
                        f'statistics for {subject_name} {str(method)}, '
                        f'{str(setting)} settings \n {traceback.format_exc()}'
                        ))
                    logger.info(
                        (f'SKIPPING {subject_name} {str(method)} '
                         f'{str(setting)} {str(statistics_type)}'))

            statistics_type = StatisticsType.SEGMENTED_TISSUES
            sheet = get_sheet_name(statistics_type, processing_pipeline,
                                   method, setting)
            logger.info(
                f'{subject_name} {str(setting)} {str(statistics_type)}')

            try:
                subject_segmented_tissues_statistics(subject_name, method_desc,
                                                     statistics_type, sheet)
            except:
                logger.exception((
                    f'Error while calculating {str(statistics_type)} '
                    f'statistics for {subject_name} {str(method)}, '
                    f'{str(setting)} settings \n {traceback.format_exc()}'))
                logger.info((
                    f'SKIPPING {subject_name} {str(method)} '
                    f'{str(statistics_type)}'))


def execute_cost_statistics(subjects: List[str], method: AnonymizationMethod,
                            processing_pipeline: ProcessingPipeline,
                            skip_default_cost: bool = False):
    """Calculates statistics for cost settings in AFNI_REFACER and PYDEFACE.

    Args:
        subjects (List[str]): subject names whose statistics will be
                        calculated, [sub-<SUBJECT_LABEL>]
        method (AnonymizationMethod): anonymization method for whose generated
                        files will be statistics calculated
        processing_pipeline (ProcessingPipeline): pipeline whose results can
                        be altered by anonymization of files
        skip_default_cost (bool, optional): if set to True, skips calculation
                        for default cost, since it was already done in default
                        statistics. Defaults to False.
    """
    cost_anonymization_settings = method.settings['cost']
    if (skip_default_cost):
        if (method.name == 'pydeface'):
            cost_anonymization_settings.remove(
                AnonymizationSetting.COST_MUTUALINFO)
        else:
            cost_anonymization_settings.remove(
                AnonymizationSetting.COST_LOCAL_PCOR_ABS)

    for cost in cost_anonymization_settings:
        execute_statistics(subjects, [method], cost, processing_pipeline)


def execute_original_vols_statistics(subjects: List[str],
                                     processing_pipeline: ProcessingPipeline
                                     ) -> None:
    ''' Calculates per voxel volume of WM, GM, CSF
    of subject's original images. '''
    statistics_type = StatisticsType.ORIGINAL_VOLS
    original_desc = MethodDescription(ORIGINAL, None, processing_pipeline,
                                      Space.SUBJECT_SPACE)

    rows = []

    for subject in subjects:
        subject_name = 'sub-' + subject
        logger.info(subject_name)

        subject_row = [subject_name]
        ok = True

        for tissue in TissueType:
            original_tissue_path = get_tissue_file_path(subject_name,
                                                        original_desc, tissue)

            if (not os.path.exists(original_tissue_path)):
                logger.exception((
                    f'Segmented tissue {str(tissue)} '
                    f'by {str(processing_pipeline)} '
                    f'does not exist for {subject_name}. '
                    f'Run first pipeline processing for {subject_name}. '
                    f'SKIPPING {subject_name}.'))
                logger.info((
                    f'Segmented tissue {str(tissue)} by '
                    f'{str(processing_pipeline)} '
                    f'does not exist for {subject_name}. '
                    f'Run first pipeline processing for {subject_name}. '
                    f'SKIPPING {subject_name}.'))
                ok = False
                break

            original_tissue = nib.load(original_tissue_path)
            vols = compute_volume(original_tissue)
            subject_row.append(vols)

        if (ok):
            rows.append(subject_row)

    df = pd.DataFrame(rows, columns=['Subject', 'WM', 'GM', 'CSF'])
    sheet = get_sheet_path(statistics_type, processing_pipeline)
    df.to_excel(sheet, sheet_name='Data', index=True)
