from enum import Enum
from typing import Union, Dict, Set


class AnonymizationSetting(Enum):
    """Enum class which represents possible anonymization settings
    which can be used when generating anonymized images.

    Raises:
        ValueError: if it can not be converted from string
    """
    DEFAULT = 'default'
    COST_MUTUALINFO = 'mutualinfo'
    COST_CORRATIO = 'corratio'
    COST_NORMCORR = 'normcorr'
    COST_NORMMI = 'normmi'
    COST_LEASTSQ = 'leastsq'
    COST_LABELDIFF = 'labeldiff'
    COST_BBR = 'bbr'
    COST_CORRATIO_MUL = 'corratio_mul'
    COST_NORM_MUTUALINFO = 'norm_mutualinfo'
    COST_HELLINGER = 'hellinger'
    COST_CORRATIO_ADD = 'corratio_add'
    COST_CORRATIO_UNS = 'corratio_uns'
    COST_LOCAL_PCOR_SIGNED = 'localPcorSigned'
    COST_LOCAL_PCOR_ABS = 'localPcorAbs'
    COST_LOCAL_PCOR_PLUS_OTHERS = 'localPcor+Others'
    COST_LOCAL_PCOR_ABS_PLUS_OTHERS = 'localPcorAbs+Others'
    BIAS_CORRECTED = 'bias_corrected'
    BRAIN_MASK_AFNI = '3dSkullStrip'

    def __init__(self, setting_name):
        self.setting_name = setting_name

    def __str__(self):
        return self.value

    @staticmethod
    def from_string(s):
        return AnonymizationSetting(s)


class Method:
    """ A class which represents method used on image
    - either we have original image or one created by
    anonymization tool.

    Attributes:
        name (str): name of method
        file_suffix (str): suffix of file which was
                            created by this method
    """
    def __init__(self, name: str, file_suffix: str):
        self.name = name
        self.file_suffix = file_suffix

    def __str__(self):
        return self.name


class AnonymizationMethod(Method):
    """ A class which represents anonymization method
    used on image.

    Attributes:
        name (str): name of method
        file_suffix (str): suffix of file which was
                            created by this method
        settings (Dict[str, Set[AnonymizationSetting]]):
                            possible settings which
                            can be used,
                            keys represent flags
        command (str): command with whom method is called
    """
    def __init__(self, name: str, filename_suffix: str,
                 settings: Dict[str, Set[AnonymizationSetting]], command: str):
        super().__init__(name, filename_suffix)
        self.settings = settings
        self.command = command

    def __str__(self):
        return self.name


class ProcessingPipeline(Enum):
    """Enum class which represents pipeline which processes data

    Raises:
        ValueError: if it can not be converted from string
    """
    SPM = 1
    FMRIPREP = 2

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_string(s):
        try:
            return ProcessingPipeline[s.upper()]
        except KeyError:
            raise ValueError()


class Space(Enum):
    """Enum class which represents space in which image is.

    Args:
        spm_file_prefix (str): prefix of file based on type of
                            spm normalization,
                            used as naming
    """
    SUBJECT_SPACE = ''
    MNI152_SPACE = 'w'
    MNI152_SPACE_ORIG_WARP = 'original-y_w'

    def __init__(self, spm_file_prefix: str):
        self.spm_file_prefix = spm_file_prefix


class MethodDescription:
    """ A class which represents how NIFTI image was processed.

    Args:
        method (AnonymizationMethod): method used for image anonymization
                                    or original
        anonymization_settings (AnonymizationSettings): settings used
                                    for anonymzation
        processing_pipeline (ProcessingPipeline): pipeline whose results
                                    are being compared
                                    if they changed due to anonymization
        normalization_type (NormalizationType): normalization_type of image
    """
    def __init__(self, method: Union[AnonymizationMethod, Method],
                 anonymization_seting: AnonymizationSetting,
                 processing_pipeline: ProcessingPipeline, space: Space):
        self.method = method
        self.anonymization_setting = anonymization_seting
        self.processing_pipeline = processing_pipeline
        self.space = space


class Mask(Enum):
    """Enum class which represents tissue masks
    used in this project to mask NIfTI images.
    """
    BRAIN = 1
    WM = 2
    GM = 3
    CSF = 4

    def __str__(self):
        return self.name


class GeneratedFile(Enum):
    """ Enum class which represents files generated by this application.
    """
    MASK = 1
    MASKED = 2
    SSIM_DIFF = 3
    OVERLAP_DIFF = 4

    def __str__(self):
        return self.name.lower()


class TissueType(Enum):
    """ Enum class which represents tissue types used in this project.

    Args:
        spm_label (str) : label prefix used in spm for describing tissue type
    """
    WM = ('c1')
    GM = ('c2')
    CSF = ('c3')

    def __init__(self, spm_label):
        self.spm_label = spm_label

    def __str__(self):
        return self.name


class StatisticsType(Enum):
    """Enum class which represents types of statistics
    calculated in this project.
    """
    MNI152_SPACE = 1
    SUBJECT_SPACE = 2
    SEGMENTED_TISSUES = 3
    MNI152_SPACE_ORIG_WARP = 4
    ORIGINAL_VOLS = 5

    def __str__(self):
        return self.name.lower()


class Metric(Enum):
    """Enum class which represents metrics used for statistics."""
    MAE = 1
    MSE = 2
    NMI = 3
    SSIM = 4
    DICE = 5

    def __str__(self):
        return self.name


class CompoundMetric:
    """ A class which represents compound metrics used for statistics.

    Args:
        metric (Metric): metric used
        mask (Mask): what part of image is analyzed
    """
    def __init__(self, metric: Metric, mask: Mask):
        self.metric = metric
        self.mask = mask

    def __str__(self):
        return self.mask.name + ' ' + self.metric.name


def create_anonymization_methods() -> Dict[str, AnonymizationMethod]:
    ''' Creates dictionary of all used anonymization methods '''
    afni_refacer_deface = AnonymizationMethod(
                'afni_refacer_deface',
                'afni_refacer_defaced',
                {'cost':
                    {AnonymizationSetting.COST_LOCAL_PCOR_ABS,
                        AnonymizationSetting.COST_LEASTSQ,
                        AnonymizationSetting.COST_MUTUALINFO,
                        AnonymizationSetting.COST_CORRATIO_MUL,
                        AnonymizationSetting.COST_NORM_MUTUALINFO,
                        AnonymizationSetting.COST_HELLINGER,
                        AnonymizationSetting.COST_CORRATIO_ADD,
                        AnonymizationSetting.COST_CORRATIO_UNS,
                        AnonymizationSetting.COST_LOCAL_PCOR_SIGNED,
                        AnonymizationSetting.COST_LOCAL_PCOR_PLUS_OTHERS,
                        AnonymizationSetting.COST_LOCAL_PCOR_ABS_PLUS_OTHERS}},
                ('@afni_refacer_run '
                    '-input {} '
                    '-mode_deface '
                    '-no_images '
                    '-prefix {}')
                )

    afni_refacer_reface = AnonymizationMethod(
                'afni_refacer_reface',
                'afni_refacer_refaced',
                {'cost':
                    {AnonymizationSetting.COST_LOCAL_PCOR_ABS,
                        AnonymizationSetting.COST_LEASTSQ,
                        AnonymizationSetting.COST_MUTUALINFO,
                        AnonymizationSetting.COST_CORRATIO_MUL,
                        AnonymizationSetting.COST_NORM_MUTUALINFO,
                        AnonymizationSetting.COST_HELLINGER,
                        AnonymizationSetting.COST_CORRATIO_ADD,
                        AnonymizationSetting.COST_CORRATIO_UNS,
                        AnonymizationSetting.COST_LOCAL_PCOR_SIGNED,
                        AnonymizationSetting.COST_LOCAL_PCOR_PLUS_OTHERS,
                        AnonymizationSetting.COST_LOCAL_PCOR_ABS_PLUS_OTHERS}},
                ('@afni_refacer_run '
                    '-input {} '
                    '-mode_reface '
                    '-no_images '
                    '-prefix {}')
                )

    fsl_deface = AnonymizationMethod(
                'fsl_deface',
                'fsl_deface_defaced',
                {'B': {AnonymizationSetting.BIAS_CORRECTED}},
                'fsl_deface {} {}'
                )

    mri_reface = AnonymizationMethod(
                'mri_reface',
                'mri_reface_refaced',
                {},
                ('mri_reface {} {} '
                    '-imType T1 '
                    '-saveQCRenders 0')
                )

    pydeface = AnonymizationMethod(
                'pydeface',
                'pydeface_defaced',
                {'cost':
                    {AnonymizationSetting.COST_MUTUALINFO,
                        AnonymizationSetting.COST_CORRATIO,
                        AnonymizationSetting.COST_NORMCORR,
                        AnonymizationSetting.COST_NORMMI,
                        AnonymizationSetting.COST_LEASTSQ,
                        AnonymizationSetting.COST_LABELDIFF,
                        AnonymizationSetting.COST_BBR}},
                ('pydeface {} '
                    '--outfile {} '
                    '--verbose '
                    '--force')
                )

    quickshear = AnonymizationMethod(
                'quickshear',
                'quickshear_defaced',
                {'afni': {AnonymizationSetting.BRAIN_MASK_AFNI}},
                'quickshear {} {} {}'
                )

    return {'afni_refacer_deface': afni_refacer_deface,
            'afni_refacer_reface': afni_refacer_reface,
            'fsl_deface': fsl_deface,
            'mri_reface': mri_reface,
            'pydeface': pydeface,
            'quickshear': quickshear}


ANONYMIZATION_METHODS = create_anonymization_methods()
ORIGINAL = Method('original', '')
